import os
import sys
import json
import requests
import urllib.parse
from conf import Configuration  # type: ignore


funcs = {"commande": {}, "action": {}, "event": {}}


class Cmd(str):
    """
    Object for text of message
    """

    webhook = "message"
    __atts = []

    def __init__(self, text):
        str.__init__(text)

    def set_atts(self, atts):
        for att in atts:
            self.__atts.append(att)

    @property
    def attachments(self):
        return self.__atts


class Payload:
    """
    Object for Payload Management
    """

    def __init__(self, payload, **kwargs) -> None:
        """
        Object for Payload Management
        """
        self.payload = payload
        self.data = kwargs

    def __str__(self):
        return self.payload

    @staticmethod
    def trt_payload_in(payload):
        """
        processing of payloads received in a sequence of structured parameters

        @params: payload [String]
        @return: payload [String] , structured parameters Dict
        """

        payload = urllib.parse.unquote(str(payload))

        res = {}
        while "{{" in payload:
            start = payload.index("{{")
            end = payload.index("}}")
            items = payload[start + 2 : end].split("===")
            res[items[0]] = items[1]
            payload = payload.replace(payload[start : end + 2], "")
        return payload.strip(), res

    @staticmethod
    def trt_payload_out(payload):
        """
        Processing of a Payload type as a character string

        @params: payload [ Payload | String ]
        @return: String
        """
        if isinstance(payload, Payload):
            tmp = ""
            for key_data, val_data in payload.data.items():
                tmp += f"{{{{{key_data}==={val_data}}}}} "
            return urllib.parse.quote(payload.payload + " " + tmp)
        return urllib.parse.quote(payload)


def analyse(data):
    """
    Function analyzing data received from Facebook
    The data received are of type Json .
    """

    def struct_atts(data):
        return data["payload"]["url"]

    for event in data["entry"]:
        messaging = event["messaging"]

        for message in messaging:

            sender_id = message["sender"]["id"]

            if message.get("message"):

                if message["message"].get("attachments"):
                    # Get file name
                    data = message["message"].get("attachments")
                    # creation de l'objet cmd personalisé
                    atts = list(map(struct_atts, data))
                    cmd = Cmd(atts[0])
                    cmd.set_atts(atts)
                    return sender_id, cmd, message
                elif message["message"].get("quick_reply"):
                    # if the response is a quick reply
                    return (
                        sender_id,
                        Cmd(message["message"]["quick_reply"].get("payload")),
                        message,
                    )
                elif message["message"].get("text"):
                    # if the response is a simple text
                    return sender_id, Cmd(message["message"].get("text")), message

            if message.get("postback"):
                recipient_id = sender_id
                pst_payload = Cmd(message["postback"]["payload"])
                pst_payload.webhook = "postback"
                return recipient_id, pst_payload, message

            if message.get("read"):
                watermark = Cmd(message["read"]["watermark"])
                watermark.webhook = "read"
                return sender_id, watermark, message

            if message.get("delivery"):
                watermark = Cmd(message["delivery"]["watermark"])
                watermark.webhook = "delivery"
                return sender_id, watermark, message

            if message.get("reaction"):
                reaction = Cmd(message["reaction"]["reaction"])
                reaction.webhook = message["reaction"]["action"]
                return sender_id, reaction, message


def command(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a processing per command sent.
    """

    def call_fn(function):
        funcs["commande"][args[0]] = function

    return call_fn


def action(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a defined action handler.
    """

    def call_fn(function):
        funcs["action"][args[0]] = function

    return call_fn


def event(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a defined event handler.
    """

    def call_fn(function):
        funcs["event"][args[0]] = function

    return call_fn


def download_file(url, file):
    """
    Downloading a file from an url.

    Args:
        url: direct link for the attachment

        file: filename with path
    """
    res = requests.get(url, allow_redirects=True)

    with open(file, "wb") as f:
        f.write(res.content)

    return file


def translate(key, lang):
    """
    translate a keyword or sentence

    @params:

        key: the key used in langs.json file

        lang: the langage code in format fr, en, mg, ...

    this function uses the langs.json file.
    """
    if not lang:
        return key

    if not os.path.isfile("langs.json"):
        print("Warning! langs.json not found", file=sys.stderr)
        from .source import langs

        with open("langs.json", "w") as fichier:
            fichier.write(langs)
            print("langs.json created!")
        return key

    with open("langs.json") as fichier:
        trans = json.load(fichier)

    keyword = trans.get(key)

    if keyword:
        if keyword.get(lang):
            return keyword.get(lang)
    return key


def simulate(sender_id, text, **params):
    """
    Simulate a message send by an user
    """
    data_json = {
        "object": "page",
        "entry": [
            {
                "messaging": [
                    {
                        "message": {
                            "text": text,
                        },
                        "sender": {"id": sender_id},
                    }
                ]
            }
        ],
    }
    header = {"content-type": "application/json; charset=utf-8"}
    return requests.post(
        f"http://127.0.0.1:{Configuration.APP_PORT}",
        json=data_json,
        headers=header,
        params=params,
    )



import os
import json
import pickle
import requests
from sys import stderr
from retry import retry
import requests_toolbelt
from .utils import Payload
from conf import Configuration  # type: ignore
from .ui import QuickReply, Button, Element


class Messenger:
    def __init__(self, log_level="error"):
        """
        Here, We need the <access token> of the facebook page we want
        to apply the bot for the purpose of page and bot interaction

        Args:
            access_token (str): A facebook page access token
            log_level (error, info, quiet): A type of display log
                error: print only error
                info: print access log and error log
                quiet: do not print anything
        """
        self.access_token = Configuration.ACCESS_TOKEN
        self.url = "https://graph.facebook.com/v13.0/me"

        if log_level not in ("error", "info", "quiet"):
            raise Exception(ValueError, "log_level must be error or info or quiet")

    def __analyse(self, res, log_level="error"):
        if log_level == "info":
            print(res.status_code, res.text)
        elif res.status_code != 200 and log_level == "error":
            print(res.status_code, res.text, file=stderr)
        return res

    @property
    def token(self):
        if self.access_token.strip() == "":
            print("Warning! EMPTY PAGE ACCESS TOKEN", file=stderr)
        return self.access_token

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_custom(self, json, endpoint="/messages"):
        """
        This method send a custom json,
        use for the api not implemented in ampalibe.

        Args:
            json: A json object
            endpoint: A endpoint of facebook api
                /messages by default like other endpoint
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + endpoint, json=json, headers=header, params=params
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_message(self, dest_id, message, prio=False):
        """
        This method allows you to send a text message to the given recipient,
        Note that the number of characters to send is limited to 2000 characters

        Args:
            dest_id (str): user id facebook for the destination
            message (str): message want to send

        Returns:
            Response: POST request to the facebook API to send a message to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#sending_text
        """
        self.send_action(dest_id, "typing_on")
        data_json = {"recipient": {"id": dest_id}, "message": {"text": message}}

        if prio:
            data_json["messaging_type"] = "MESSAGE_TAG"
            data_json["tag"] = "ACCOUNT_UPDATE"

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages", json=data_json, headers=header, params=params
        )
        self.send_action(dest_id, "typing_off")
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_action(self, dest_id, action):
        """
        This method is used to simulate an action on messages.
        example: view, writing.
        Action available: ['mark_seen', 'typing_on', 'typing_off']

        Args:
            dest_id (str): user id facebook for the destination
            action (str): action ['mark_seen', 'typing_on', 'typing_off']

        Returns:
            Response: POST request to the facebook API to send an action to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/sender-actions
        """
        data_json = {
            "messaging_type": "RESPONSE",
            "recipient": {"id": dest_id},
            "sender_action": action,
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages", json=data_json, headers=header, params=params
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_quick_reply(self, dest_id, quick_rep, text, next=False):
        """
        Quick replies provide a way to present a set of up to 13 buttons
        in-conversation that contain a title and optional image, and appear
        prominently above the composer. You can also use quick replies
        to request a person's location, email address, and phone number.

        Args:
            dest_id (str): user id facebook for the destination
            quick_rep (list of dict) || (list of QuickRep): list of the different quick_reply to send a user
            text (str): A text of a little description for each <quick_reply>
            next(bool) || (text): this params activate the next page when elements have a length more than ten

        Returns:
            Response: POST request to the facebook API to send a quick_reply to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/quick-replies
        """

        for i in range(len(quick_rep[:12])):
            if isinstance(quick_rep[i], QuickReply):
                quick_rep[i] = quick_rep[i].value
            if quick_rep[i].get("payload"):
                quick_rep[i]["payload"] = Payload.trt_payload_out(
                    quick_rep[i]["payload"]
                )

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": {"id": dest_id},
            "message": {"text": text, "quick_replies": quick_rep[:12]},
        }

        if len(quick_rep) > 13 and next:
            dataJSON["message"]["quick_replies"].append(
                {
                    "content_type": "text",
                    "title": "More" if next == True else str(next),
                    "payload": "/__more",
                    "image_url": "https://icon-icons.com/downloadimage.php"
                    + "?id=81300&root=1149/PNG/512/&file="
                    + "1486504364-chapter-controls-forward-play"
                    + "-music-player-video-player-next_81300.png",
                }
            )
            pickle.dump(
                (quick_rep[13:], text, next), open(f"assets/private/.__{dest_id}", "wb")
            )
        else:
            if len(quick_rep) > 12:
                quick_rep[12]["payload"] = Payload.trt_payload_out(
                    quick_rep[12]["payload"]
                )
                dataJSON["message"]["quick_replies"].append(quick_rep[12])

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages", json=dataJSON, headers=header, params=params
        )
        return self.__analyse(res)

    def send_result(self, **ext):
        raise Exception("Deprecated: Use send_template instead!")

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_template(self, dest_id, elements, quick_rep=None, next=None):
        """
        The method represent a Message templates who offer a way for you
        to offer a richer in-conversation experience than standard text messages by integrating
        buttons, images, lists, and more alongside text a single message. Templates can be use for
        many purposes, such as displaying product information, asking the messagerecipient to choose
        from a pre-determined set of options, and showing search results.

        For this, messenger only validates 10 templates
        for the first display, so we put the parameter
        <next> to manage these numbers if it is a number of
        elements more than 10,

        So, there is a quick_reply which acts as a "next page"
        displaying all requested templates

        Args:
            dest_id (str): user id facebook for the destination
            elements (list of dict) || (list of Element): the list of the specific elements to define the
            structure for the template
            quick_rep (list of dict) || (list of QuickReply): addition quick reply at the bottom of the template
            next(bool) || (text): this params activate the next page when elements have a length more than ten

        Returns:
            Response: POST request to the facebook API to send a template generic to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/template/generic
        """

        for i in range(len(elements[:10])):
            if isinstance(elements[i], Element):
                elements[i] = elements[i].value
            for j in range(len(elements[i]["buttons"])):
                if elements[i]["buttons"][j].get("payload"):
                    elements[i]["buttons"][j]["payload"] = Payload.trt_payload_out(
                        elements[i]["buttons"][j]["payload"]
                    )

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": {"id": dest_id},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "generic",
                        "elements": elements[:10],
                    },
                },
            },
        }

        if len(elements) > 10 and next:
            dataJSON["message"]["quick_replies"] = [
                {
                    "content_type": "text",
                    "title": "Next" if next == True else str(next),
                    "payload": "/__next",
                    "image_url": "https://icon-icons.com/downloadimage.php"
                    + "?id=81300&root=1149/PNG/512/&file="
                    + "1486504364-chapter-controls-forward-play"
                    + "-music-player-video-player-next_81300.png",
                }
            ]
            pickle.dump(
                (elements[10:], next), open(f"assets/private/.__{dest_id}", "wb")
            )

        if quick_rep:
            for i in range(len(quick_rep)):
                if isinstance(quick_rep[i], QuickReply):
                    quick_rep[i] = quick_rep[i].value
            dataJSON["message"]["quick_replies"] = quick_rep

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages", json=dataJSON, headers=header, params=params
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_file_url(self, dest_id, url, filetype="file"):
        """
        The Messenger Platform allows you to attach assets to messages, including audio,
        video, images, and files.All this is the role of this Method. The maximum attachment
        size is 25 MB. The maximum resolution for images is 85 Megapixel. There are three ways
        to attach an asset to a message:

        Args:
            dest_id (str): user id facebook for destination
            url (str): the origin url for the file
            filetype (str, optional): type of the file["video","image",...]. Defaults to 'file'.

        Returns:
            Response: POST request to the facebook API to send a template generic to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#url
        """

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": {"id": dest_id},
            "message": {
                "attachment": {
                    "type": filetype,
                    "payload": {"url": url, "is_reusable": True},
                }
            },
        }
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}
        res = requests.post(
            self.url + "/messages", json=dataJSON, headers=header, params=params
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def persistent_menu(self, dest_id, persistent_menu, action="PUT", **kwargs):
        """
        The Persistent Menu disabling the composer best practices allows you to have an always-on
        user interface element inside Messenger conversations. This is an easy way to help people
        discover and access the core functionality of your Messenger bot at any point in the conversation

        Args:
            dest_id (str): user id for destination
            persistent_menu (list of dict) || (list of Button): the elements of the persistent menu to enable
            action (str, optional): the action for benefit["PUT","DELETE"]. Defaults to 'PUT'.

            locale [optionnel]
            composer_input_disabled [optionnel]

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/persistent-menu
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        persistent_menu = [
            persistent.value if isinstance(persistent, Button) else persistent
            for persistent in persistent_menu
        ]

        if action == "PUT":
            dataJSON = {
                "psid": dest_id,
                "persistent_menu": [
                    {
                        "locale": kwargs.get("locale", "default"),
                        "composer_input_disabled": kwargs.get(
                            "composer_input_disabled", "false"
                        ),
                        "call_to_actions": persistent_menu,
                    }
                ],
            }

            res = requests.post(
                self.url + "/custom_user_settings",
                json=dataJSON,
                headers=header,
                params=params,
            )
            return self.__analyse(res)

        elif action == "DELETE":
            params["params"] = "(persistent_menu)"
            params["psid"] = dest_id

            res = requests.delete(
                self.url + "/custom_user_settings", headers=header, params=params
            )
            return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_file(self, dest_id, file, filetype="file", filename=None):
        """
        this method send an attachment from file

        Args:
            destId (str): user id facebook for the destination
            file (str): name of the file in local folder
            filetype (str, optional): type of the file["video","image",...]. Defaults to "file".
            filename (str, optional): A filename received for de destination . Defaults to None.

        Returns:
            Response: POST request to the facebook API to send a file to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#file
        """
        if filename is None:
            filename = file
        params = {"access_token": self.token}

        data = {
            "recipient": json.dumps({"id": dest_id}),
            "message": json.dumps(
                {
                    "attachment": {
                        "type": filetype,
                        "payload": {
                            "is_reusable": True,
                        },
                    }
                }
            ),
            "filedata": (
                os.path.basename(filename),
                open(file, "rb"),
                f"{filetype}/{file.split('.')[-1]}",
            ),
        }

        # multipart encode the entire payload
        multipart_data = requests_toolbelt.MultipartEncoder(data)

        # multipart header from multipart_data
        header = {"Content-Type": multipart_data.content_type}

        res = requests.post(
            self.url + "/messages", params=params, headers=header, data=multipart_data
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_media(self, dest_id, fb_url, media_type):
        """
        Method that sends files media as image and video via facebook link.
        This model does not allow any external URLs, only those on Facebook.

        Args:
            destId (str): user id facebook for the destination
            fb_url (str): url of the media to send on facebook
                for this: To get the Facebook URL of an image or video, follow these steps:
                    -Click on the image or video thumbnail to open the full-size view
                    -Copy the URL address from your browser's address bar.
            media_type (str): the type of the media who to want send, available["image","video"]

        Returns:
            Response: POST request to the facebook API to send a media file using url facebook

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/template/media
        """
        dataJSON = {
            "recipient": {"id": dest_id},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "media",
                        "elements": [{"media_type": media_type, "url": fb_url}],
                    },
                }
            },
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            "https://graph.facebook.com/v6.0/me/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def get_started(self, payload="/"):
        """
        Method that GET STARTED button
        when the user talk first to the bot.

        Returns:
            Response: POST request to the facebook API to send a media file using url facebook

        Ref:
            https://developers.facebook.com/docs/messenger-platform/reference/messenger-profile-api/get-started-button
        """

        dataJSON = {"get_started": {"payload": payload}}

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            "https://graph.facebook.com/v6.0/me/messenger_profile",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_button(self, dest_id, buttons, text):
        """
        The button template sends a text message with
        up to three buttons attached. This template gives
        the message recipient different options to choose from,
        such as predefined answers to questions or actions to take.

         Args:
             dest_id (str): user id facebook for the destination
             buttons (list of dict) || (list of Button): The list of buttons who want send
             text (str): A text to describe the fonctionnality of the buttons

         Returns:
             Response: POST request to the facebook API to send all different buttons

         Ref:
             https://developers.facebook.com/docs/messenger-platform/send-messages/template/button
        """

        for i in range(len(buttons)):
            if isinstance(buttons[i], Button):
                buttons[i] = buttons[i].value
            if buttons[i].get("payload"):
                buttons[i]["payload"] = Payload.trt_payload_out(buttons[i]["payload"])

        self.send_action(dest_id, "typing_on")
        data_json = {
            "recipient": {"id": dest_id},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": text,
                        "buttons": buttons,
                    },
                }
            },
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        self.send_action(dest_id, "typing_off")
        res = requests.post(
            self.url + "/messages", json=data_json, headers=header, params=params
        )
        return self.__analyse(res)



"""
    List of All UI Widget Messenger 
"""


class QuickReply:
    def __init__(self, **kwargs):
        """
        Object that can be used to generated a quick_reply
        """
        self.content_type = kwargs.get("content_type", "text")
        if self.content_type not in ("text", "user_phone_number", "user_email"):
            raise ValueError(
                "content_type can only be 'text', 'user_phone_number', 'user_email'"
            )

        self.title = kwargs.get("title")
        self.payload = kwargs.get("payload")

        if self.content_type == "text" and not self.payload:
            raise ValueError("payload must be present for text")

        if self.content_type == "text" and not self.title:
            raise ValueError("title must be present for text")

        self.image_url = kwargs.get("image_url")

    @property
    def value(self):
        res = {"content_type": self.content_type}

        if self.content_type == "text":
            res["title"] = self.title
            res["payload"] = self.payload

            if self.image_url:
                res["image_url"] = self.image_url
        return res

    def __str__(self):
        return str(self.value)


class Button:
    def __init__(self, **kwargs):

        self.type = kwargs.get("type", "postback")
        self.title = kwargs.get("title")
        self.url = kwargs.get("url")
        self.payload = kwargs.get("payload")

        if self.type not in (
            "postback",
            "web_url",
            "phone_number",
            "account_link",
            "account_unlink",
        ):
            raise ValueError(
                "type must be one of 'postback', 'web_url', 'phone_number', 'account_link', 'account_unlink'"
            )

        if self.type in ("postback", "phone_number"):

            if not self.payload:
                raise ValueError("payload must be present")

            if not self.title:
                raise ValueError("title must be present")

        elif self.type == "web_url":
            if not self.url:
                raise ValueError("url must be present for web_url")

            if not self.title:
                raise ValueError("title must be present for web_url")

        elif self.type == "account_link":
            if not self.url:
                raise ValueError("url facebook login must be present for account_link")

    @property
    def value(self):

        if self.type in ("postback", "phone_number"):
            return {"type": self.type, "title": self.title, "payload": self.payload}

        elif self.type == "web_url":
            return {"type": "web_url", "title": self.title, "url": self.url}

        elif self.type == "account_link":
            return {"type": "account_link", "url": self.url}

        elif self.type == "account_unlink":
            return {"type": "account_unlink"}

    def __str__(self):
        return str(self.value)


class Element:
    def __init__(self, **kwargs):
        self.title = kwargs.get("title")
        self.subtitle = kwargs.get("subtitle")
        self.image = kwargs.get("image_url")
        self.buttons = kwargs.get("buttons")

        if not self.title:
            raise ValueError("Element must be have a title")

        if not self.buttons:
            raise ValueError("Element must be have a buttons")

        if not isinstance(self.buttons, list):
            raise ValueError("buttons must be a list of Button")

        if len(self.buttons) > 3:
            raise ValueError("buttons must be three maximum")

        for i in range(len(self.buttons)):
            if not isinstance(self.buttons[i], Button):
                raise ValueError("buttons must a List of Button")

    @property
    def value(self):
        res = {"title": self.title}

        if self.subtitle:
            res["subtitle"] = self.subtitle

        if self.image:
            res["image_url"] = self.image

        res["buttons"] = [button.value for button in self.buttons]

        return res

    def __str__(self):
        return str(self.value)



import json
from conf import Configuration  # type: ignore


class Model:
    """
    Object for interact with database with pre-defined function
    """

    def __init__(self, conf=Configuration):
        """
        object to interact with database

        @params: conf [ Configuration object ]
        @return: Request object
        """
        self.ADAPTER = conf.ADAPTER
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            self.DB_CONF = {
                "host": conf.DB_HOST,
                "user": conf.DB_USER,
                "password": conf.DB_PASSWORD,
                "database": conf.DB_NAME,
                "port": conf.DB_PORT,
            }
        else:  # SQLite is choosen by default
            self.DB_CONF = conf.DB_FILE

        self.__connect()
        self.__init_db()

    def __connect(self):
        """
        The function which connect object to the database.
        """
        if self.ADAPTER == "MYSQL":
            import mysql.connector

            self.db = mysql.connector.connect(**self.DB_CONF)
        elif self.ADAPTER == "POSTGRESQL":
            import psycopg2

            self.db = psycopg2.connect(**self.DB_CONF)
        else:
            import sqlite3

            self.db = sqlite3.connect(self.DB_CONF)

        self.cursor = self.db.cursor()

    def __init_db(self):
        """
        Creation of table if not exist
        Check the necessary table if exists
        """

        if self.ADAPTER == "MYSQL":
            req = """
                CREATE TABLE IF NOT EXISTS `amp_user` (
                    `id` INT NOT NULL AUTO_INCREMENT,
                    `user_id` varchar(50) NOT NULL UNIQUE,
                    `action` varchar(50) DEFAULT NULL,
                    `last_use` datetime NOT NULL DEFAULT current_timestamp(),
                    `lang` varchar(5) DEFAULT NULL,
                    `tmp` varchar(255) DEFAULT NULL,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;
            """
        elif self.ADAPTER == "POSTGRESQL":
            req = """
                CREATE TABLE IF NOT EXISTS  "amp_user" (
                    id SERIAL,
                    user_id VARCHAR NULL DEFAULT NULL,
                    action VARCHAR NULL DEFAULT NULL,
                    tmp VARCHAR NULL DEFAULT NULL,
                    last_use TIMESTAMP NULL DEFAULT NOW(),
                    lang VARCHAR NULL DEFAULT NULL,
                    PRIMARY KEY (id),
                    UNIQUE (user_id)
                )
            """
        else:
            req = """
               CREATE TABLE IF NOT EXISTS amp_user (
                   id INTEGER PRIMARY KEY AUTOINCREMENT, 
                   user_id TEXT NOT NULL UNIQUE,
                   action TEXT,
                   last_use TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                   lang TEXT,
                   tmp TEXT
                )
            """
        self.cursor.execute(req)
        self.db.commit()

    def verif_db(fonction):
        """
        decorator that checks if the database
        is connected or not before doing an operation.
        """

        def trt_verif(*arg, **kwarg):
            if arg[0].ADAPTER == "MYSQL":
                if not arg[0].db.is_connected():
                    # reconnexion de la base
                    try:
                        arg[0].db.reconnect()
                    except Exception:
                        arg[0].__connect()
            elif arg[0].ADAPTER == "POSTGRESQL":
                try:
                    arg[0].cursor.execute("SELECT 1")
                except Exception:
                    arg[0].__connect()
            else:
                arg[0].__connect()
            return fonction(*arg, **kwarg)

        return trt_verif

    @verif_db
    def _verif_user(self, user_id):
        """
        method to insert new user and/or update the date
        of last use if the user already exists.

        @params :  user_id

        """
        # Insertion dans la base si non present
        # Mise à jour du last_use si déja présent
        if self.ADAPTER == "MYSQL":
            req = """
                INSERT INTO amp_user(user_id) VALUES (%s)
                ON DUPLICATE KEY UPDATE last_use = NOW()
            """
        elif self.ADAPTER == "POSTGRESQL":
            req = """
                INSERT INTO amp_user(user_id) VALUES (%s)
                ON CONFLICT (user_id) DO UPDATE SET last_use = NOW();
            """
        else:
            req = """
                INSERT INTO amp_user(user_id) VALUES (?)
                ON CONFLICT(user_id) DO UPDATE SET last_use = CURRENT_TIMESTAMP;
            """
        self.cursor.execute(req, (user_id,))
        self.db.commit()

    @verif_db
    def get_action(self, user_id):
        """
        get current action of an user

         @params :  user_id
         @return : current action [ type of String/None ]
        """
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "SELECT action FROM amp_user WHERE user_id = %s"
        else:
            req = "SELECT action FROM amp_user WHERE user_id = ?"
        self.cursor.execute(req, (user_id,))
        # retourne le resultat
        return self.cursor.fetchone()[0]

    @verif_db
    def set_action(self, user_id, action):
        """
        define a current action if an user

        @params :  user_id
        @return:  None
        """
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "UPDATE amp_user set action = %s WHERE user_id = %s"
        else:
            req = "UPDATE amp_user set action = ? WHERE user_id = ?"
        self.cursor.execute(req, (action, user_id))
        self.db.commit()

    @verif_db
    def __get_temp(self, user_id):
        """
        get all temporary data of an user

        @params :  user_id
        @return: JSON string
        """
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "SELECT tmp FROM amp_user WHERE user_id = %s"
        else:
            req = "SELECT tmp FROM amp_user WHERE user_id = ?"
        self.cursor.execute(req, (user_id,))
        res = self.cursor.fetchone()
        return res if res is None else res[0]

    @verif_db
    def set_temp(self, user_id, key, value):
        """
        set a temp parameter of an user

         @params:  user_id
         @return:  None
        """
        data = self.__get_temp(user_id)
        if not data:
            data = {}
        else:
            data = json.loads(data)
        data[key] = value
        data = json.dumps(data)
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "UPDATE amp_user SET tmp = %s WHERE user_id = %s"
        else:
            req = "UPDATE amp_user SET tmp = ? WHERE user_id = ?"
        self.cursor.execute(req, (data, user_id))
        self.db.commit()

    @verif_db
    def get_temp(self, user_id, key):
        """
        get one temporary data of an user

        @parmas :  user_id
                   key
        @return: data
        """
        data = self.__get_temp(user_id)
        if not data:
            return
        data = json.loads(data)
        return data.get(key)

    @verif_db
    def del_temp(self, user_id, key):
        """
        delete temporary parameter of an user

        @parameter :  user_id
                      key
        @return: None
        """
        data = self.__get_temp(user_id)
        if not data:
            return
        data = json.loads(data)
        try:
            data.pop(key)
        except KeyError:
            pass
        else:
            data = json.dumps(data)
            if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
                req = "UPDATE amp_user SET tmp = %s WHERE user_id = %s"
            else:
                req = "UPDATE amp_user SET tmp = ? WHERE user_id = ?"
            self.cursor.execute(req, (data, user_id))
            self.db.commit()

    @verif_db
    def get_lang(self, user_id):
        """
        get current lang of an user

        @params: user_id
        @return lang or None
        """
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "SELECT lang FROM amp_user WHERE user_id = %s"
        else:
            req = "SELECT lang FROM amp_user WHERE user_id = ?"
        self.cursor.execute(req, (user_id,))
        return self.cursor.fetchone()[0]

    @verif_db
    def set_lang(self, user_id, lang):
        """
        define a current lang for an user

        @params :  user_id
        @return:  None
        """
        if self.ADAPTER == "MYSQL" or self.ADAPTER == "POSTGRESQL":
            req = "UPDATE amp_user set lang = %s WHERE user_id = %s"
        else:
            req = "UPDATE amp_user set lang = ? WHERE user_id = ?"
        self.cursor.execute(req, (lang, user_id))
        self.db.commit()



import os
import sys
import pickle
import asyncio
import uvicorn
from .model import Model
from threading import Thread
from conf import Configuration  # type: ignore
from .messenger import Messenger
from fastapi.staticfiles import StaticFiles
from fastapi import FastAPI, Request, Response
from .utils import funcs, analyse, Cmd, Payload

_req = None
loop = None

webserver = FastAPI(title="Ampalibe server")
if os.path.isdir("assets/public"):
    webserver.mount("/asset", StaticFiles(directory="assets/public"), name="asset")


class Extra:
    def __init__(self, *args):
        self.query = Model()
        self.chat = Messenger()

    @staticmethod
    def run():
        """
        function that run framework
        """
        global _req
        _req = Model()

        global loop
        loop = asyncio.get_event_loop()
        Thread(target=loop.run_forever).start()

        uvicorn.run(webserver, port=Configuration.APP_PORT, host=Configuration.APP_HOST)


class Server:
    """
    Content of webhook
    """

    @webserver.on_event("shutdown")
    def shutdow():
        """
        function that shutdown crontab server
        """
        loop.call_soon_threadsafe(loop.stop)

    @webserver.get("/")
    async def verif(request: Request):
        """
        Main verification for bot server is received here
        """
        fb_token = request.query_params.get("hub.verify_token")

        if fb_token == Configuration.VERIF_TOKEN:
            return Response(content=request.query_params["hub.challenge"])
        return "Failed to verify token"

    @webserver.post("/")
    async def main(request: Request):
        """
        Main Requests for bot messenger is received here.
        """
        testmode = request.query_params.get("testmode")
        data = await request.json()

        # data analysis and decomposition
        sender_id, payload, message = analyse(data)

        if payload.webhook not in ("message", "postback"):
            if funcs["event"].get(payload.webhook):
                kw = {"sender_id": sender_id, "watermark": payload, "message": message}
                if testmode:
                    funcs["event"][payload.webhook](**kw)
                else:
                    Thread(target=funcs["event"][payload.webhook], kwargs=kw).start()
            return {"status": "ok"}

        _req._verif_user(sender_id)
        # get action for the current user
        action = _req.get_action(sender_id)
        lang = _req.get_lang(sender_id)

        if payload in ("/__next", "/__more"):
            bot = Messenger()
            if os.path.isfile(f"assets/private/.__{sender_id}"):
                elements = pickle.load(open(f"assets/private/.__{sender_id}", "rb"))

                if payload == "/__next":
                    bot.send_template(sender_id, elements[0], next=elements[1])
                else:
                    bot.send_quick_reply(
                        sender_id, elements[0], elements[1], next=elements[2]
                    )

                return {"status": "ok"}

        if os.path.isfile(f"assets/private/.__{sender_id}"):
            os.remove(f"assets/private/.__{sender_id}")

        if action and funcs["action"].get(action):
            """
            CASE an action is set.
            """
            kw = {
                "sender_id": sender_id,
                "cmd": payload,
                "message": message,
                "lang": lang,
            }

            if testmode:
                funcs["action"].get(action)(**kw)
            else:
                Thread(target=funcs["action"].get(action), kwargs=kw).start()
        else:
            if action:
                print(f'Warning! action: "{action}" Not found', file=sys.stderr)
            payload, kw = Payload.trt_payload_in(payload)
            kw["sender_id"] = sender_id
            kw["cmd"] = Cmd(payload)  # Remake payload to CMD object
            kw["message"] = message
            kw["lang"] = lang
            if testmode:
                funcs["commande"].get(payload.split()[0], funcs["commande"]["/"])(**kw)
            else:
                Thread(
                    target=funcs["commande"].get(
                        payload.split()[0], funcs["commande"]["/"]
                    ),
                    kwargs=kw,
                ).start()

        return {"status": "ok"}
